/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998, 1999, 2000  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_ph_events.c,v 1.1.2.2 2000/12/22 04:32:48 hercules Exp $";
#endif

/* Handle the event stream, converting photon events into SDL events */

#include <Ph.h>
#include <stdio.h>
#include <setjmp.h>
#include <photon/PkKeyDef.h>
#include <sys/time.h>

#include "SDL.h"
#include "SDL_syswm.h"
#include "SDL_sysevents.h"
#include "SDL_sysvideo.h"
#include "SDL_events_c.h"
#include "SDL_ph_video.h"
#include "SDL_ph_modes_c.h"
#include "SDL_ph_image_c.h"
#include "SDL_ph_events_c.h"


/* The translation tables from a photon keysym to a SDL keysym */
static SDLKey ODD_keymap[256];
static SDLKey MISC_keymap[0xFFFF + 1];
SDL_keysym *ph_TranslateKey(PhKeyEvent_t *key, SDL_keysym *keysym);

/* Check to see if this is a repeated key.
   (idea shamelessly lifted from GII -- thanks guys! :)
 */
/*
static int X11_KeyRepeat(Display *display, XEvent *event)
{
	XEvent peekevent;
	int repeated;

	repeated = 0;
	if ( XPending(display) ) {
		XPeekEvent(display, &peekevent);
		if ( (peekevent.type == KeyPress) &&
		     (peekevent.xkey.keycode == event->xkey.keycode) &&
		     (peekevent.xkey.time == event->xkey.time) ) {
			repeated = 1;
			XNextEvent(display, &peekevent);
		}
	}
	return(repeated);
}
*/
/* Note:  The X server buffers and accumulates mouse motion events, so
   the motion event generated by the warp may not appear exactly as we
   expect it to.  We work around this (and improve performance) by only
   warping the pointer when it reaches the edge, and then wait for it.
*/
/*
#define MOUSE_FUDGE_FACTOR	8

static inline int X11_WarpedMotion(_THIS, XEvent *xevent)
{
	int w, h, i;
	int deltax, deltay;
	int posted;

	w = SDL_VideoSurface->w;
	h = SDL_VideoSurface->h;
	deltax = xevent->xmotion.x - mouse_last.x;
	deltay = xevent->xmotion.y - mouse_last.y;
#ifdef DEBUG_MOTION
  printf("Warped mouse motion: %d,%d\n", deltax, deltay);
#endif
	mouse_last.x = xevent->xmotion.x;
	mouse_last.y = xevent->xmotion.y;
	posted = SDL_PrivateMouseMotion(0, 1, deltax, deltay);

	if ( (xevent->xmotion.x < MOUSE_FUDGE_FACTOR) ||
	     (xevent->xmotion.x > (w-MOUSE_FUDGE_FACTOR)) ||
	     (xevent->xmotion.y < MOUSE_FUDGE_FACTOR) ||
	     (xevent->xmotion.y > (h-MOUSE_FUDGE_FACTOR)) ) {
		/* Get the events that have accumulated */
/*		while ( XCheckTypedEvent(SDL_Display, MotionNotify, xevent) ) {
			deltax = xevent->xmotion.x - mouse_last.x;
			deltay = xevent->xmotion.y - mouse_last.y;
#ifdef DEBUG_MOTION
  printf("Extra mouse motion: %d,%d\n", deltax, deltay);
#endif
			mouse_last.x = xevent->xmotion.x;
			mouse_last.y = xevent->xmotion.y;
			posted += SDL_PrivateMouseMotion(0, 1, deltax, deltay);
		}
		mouse_last.x = w/2;
		mouse_last.y = h/2;
		XWarpPointer(SDL_Display, None, SDL_Window, 0, 0, 0, 0,
					mouse_last.x, mouse_last.y);
		for ( i=0; i<10; ++i ) {
        		XMaskEvent(SDL_Display, PointerMotionMask, xevent);
			if ( (xevent->xmotion.x >
			          (mouse_last.x-MOUSE_FUDGE_FACTOR)) &&
			     (xevent->xmotion.x <
			          (mouse_last.x+MOUSE_FUDGE_FACTOR)) &&
			     (xevent->xmotion.y >
			          (mouse_last.y-MOUSE_FUDGE_FACTOR)) &&
			     (xevent->xmotion.y <
			          (mouse_last.y+MOUSE_FUDGE_FACTOR)) ) {
				break;
			}
#ifdef DEBUG_XEVENTS
  printf("Lost mouse motion: %d,%d\n", xevent->xmotion.x, xevent->xmotion.y);
#endif
		}
#ifdef DEBUG_XEVENTS
		if ( i == 10 ) {
			printf("Warning: didn't detect mouse warp motion\n");
		}
#endif
	}
	return(posted);
}
*/

static int ph_DispatchEvent(_THIS)
{
	int posted;
	PhRect_t* rect;
	PhPointerEvent_t* pointerEvent;
	PhKeyEvent_t* keyEvent;
	PhWindowEvent_t* winEvent;

	posted = 0;
	switch (event->type) {
		case Ph_EV_BOUNDARY:
		{
			if (event->subtype == Ph_EV_PTR_ENTER)
				posted = SDL_PrivateAppActive(1, SDL_APPMOUSEFOCUS);
			else if (event->subtype ==Ph_EV_PTR_LEAVE)
				posted = SDL_PrivateAppActive(0, SDL_APPMOUSEFOCUS);	
		}
		break;

		case Ph_EV_PTR_MOTION_BUTTON:
		case Ph_EV_PTR_MOTION_NOBUTTON:
		{
			if ( SDL_VideoSurface ) {
				pointerEvent = PhGetData( event );
				rect = PhGetRects( event );
				posted = SDL_PrivateMouseMotion(0, 1,
                		pointerEvent->pos.x - rect[0].ul.x,
                		pointerEvent->pos.y - rect[0].ul.y);		
			}
		}
		break;

		case Ph_EV_BUT_PRESS:
		{
			pointerEvent = PhGetData( event );
			/* TODO: is 'buttons' the right mapping? */
			posted = SDL_PrivateMouseButton(SDL_PRESSED,
                        	pointerEvent->buttons, 0, 0);
		}
		break;

		case Ph_EV_BUT_RELEASE:
		{
			pointerEvent = PhGetData( event );
			 posted = SDL_PrivateMouseButton(SDL_RELEASED,
         			pointerEvent->buttons, 0, 0);
		}
		break;

		case Ph_EV_WM:
		{
			winEvent = PhGetData( event );
			
			/* losing focus */
			if ((winEvent->event_f==Ph_WM_FOCUS)&&
				(winEvent->event_state==Ph_WM_EVSTATE_FOCUSLOST))
			{
				posted = SDL_PrivateAppActive(0, SDL_APPINPUTFOCUS);	

				/* Queue leaving fullscreen mode */
				switch_waiting = 0x01;
				switch_time = SDL_GetTicks() + 200;
			}

			/* gaining focus */
			else if ((winEvent->event_f==Ph_WM_FOCUS)&&
					(winEvent->event_state==Ph_WM_EVSTATE_FOCUS))
			{
				posted = SDL_PrivateAppActive(1, SDL_APPINPUTFOCUS);

				/* Queue entry into fullscreen mode */
				switch_waiting = 0x01 | SDL_FULLSCREEN;
				switch_time = SDL_GetTicks() + 1500;
			}

			/* request to quit */
			else if (winEvent->event_f==Ph_WM_CLOSE)
			{
				posted = SDL_PrivateQuit();
			}
		}
		break;
		
		/* window has been resized, moved or removed */
		case Ph_EV_EXPOSE:
		{
			if (SDL_VideoSurface)
			{
				rect = PhGetRects( event );
				PgSetClipping(1, rect );
			}
		}
		break;

		case Ph_EV_KEY:
		{
			SDL_keysym keysym;
			keyEvent = PhGetData( event );

			if (Pk_KF_Key_Down & keyEvent->key_flags)
			{
				posted = SDL_PrivateKeyboard(SDL_PRESSED,
                			ph_TranslateKey(&keyEvent, &keysym));
			}
			else /* must be key release */
 			{
				posted = SDL_PrivateKeyboard(SDL_PRESSED,
        				ph_TranslateKey(&keyEvent, &keysym));
			}
		}
		break;
	}
	return(posted);
}

/* perform a blocking read if no events available */
int ph_Pending(_THIS)
{
	int fl;

	/* Flush the display connection and look to see if events are queued */
//	XFlush(display);
	PgFlush();

	do {

	    fl=PhEventNext( event,EVENT_SIZE );

	} while( fl!=Ph_EVENT_MSG );

//	if ( XEventsQueued(display, QueuedAlready) ) {

	/* Oh well, nothing is ready .. */
	return(0);
}

/*
SAMPLE EVENT PUMP
=================
static void update( int block ){

    int ch,fl;
    PhKeyEvent_t *key;

    for( ;; ){

        if( block ){
            do{
                fl=PhEventNext( event,EVENT_SIZE );
            }while( fl!=Ph_EVENT_MSG );
            block=0;
        }else{
            do{
                fl=PhEventPeek( event,EVENT_SIZE );
                if( !fl ) return;
            }while( fl!=Ph_EVENT_MSG );
        }

        switch( event->type ){
        case Ph_EV_KEY:
            key=PhGetData( event );
            ch=key->key_cap;    // & 127;
            fl=key->key_flags;
            if( ch<32 || ch>127 ) break;
            if( fl & Pk_KF_Key_Down ){
                if( !(fl & Pk_KF_Key_Repeat) ){
                    if( queput-queget<QUE_SIZE ) keyque[ queput++ & QUE_MASK ]=ch;
                    keyMatrix[ch]=1;
                }
            }else{
                keyMatrix[ch]=0;
            }
            break;
        default:
            PtEventHandler( event );
        }
    }
}
*/

void ph_PumpEvents(_THIS)
{
	int pending;

	/* Keep processing pending events */
	pending = 0;
	while ( ph_Pending(this) ) {
		ph_DispatchEvent(this);
		++pending;
	}
	if ( switch_waiting ) {
		Uint32 now;

		now  = SDL_GetTicks();
		if ( pending || !SDL_VideoSurface ) {
			/* Try again later... */
			if ( switch_waiting & SDL_FULLSCREEN ) {
				switch_time = now + 1500;
			} else {
				switch_time = now + 200;
			}
		} else if ( now >= switch_time ) {
			Uint32 go_fullscreen;

			go_fullscreen = switch_waiting & SDL_FULLSCREEN;
			switch_waiting = 0;
			if ( SDL_VideoSurface->flags & SDL_FULLSCREEN ) {
				if ( go_fullscreen ) {
					ph_EnterFullScreen(this);
				} else {
					ph_LeaveFullScreen(this);
				}
			}
			/* Handle focus in/out when grabbed */
/*
			if ( go_fullscreen ) {
				ph_GrabInputNoLock(this, this->input_grab);
			} else {
				ph_GrabInputNoLock(this, SDL_GRAB_OFF);
			}
*/
		}
	}
}

void ph_InitKeymap(void)
{
	int i;

	/* Odd keys used in international keyboards */
	for ( i=0; i<SDL_TABLESIZE(ODD_keymap); ++i )
		ODD_keymap[i] = SDLK_UNKNOWN;

	/* Map the miscellaneous keys */
	for ( i=0; i<SDL_TABLESIZE(MISC_keymap); ++i )
		MISC_keymap[i] = SDLK_UNKNOWN;

	MISC_keymap[Pk_BackSpace] = SDLK_BACKSPACE;
	MISC_keymap[Pk_Tab] = SDLK_TAB;
	MISC_keymap[Pk_Clear] = SDLK_CLEAR;
	MISC_keymap[Pk_Return] = SDLK_RETURN;
	MISC_keymap[Pk_Pause] = SDLK_PAUSE;
	MISC_keymap[Pk_Escape] = SDLK_ESCAPE;
	MISC_keymap[Pk_Delete] = SDLK_DELETE;

	MISC_keymap[Pk_KP_0] = SDLK_KP0;
	MISC_keymap[Pk_KP_1] = SDLK_KP1;
	MISC_keymap[Pk_KP_2] = SDLK_KP2;
	MISC_keymap[Pk_KP_3] = SDLK_KP3;
	MISC_keymap[Pk_KP_4] = SDLK_KP4;
	MISC_keymap[Pk_KP_5] = SDLK_KP5;
	MISC_keymap[Pk_KP_6] = SDLK_KP6;
	MISC_keymap[Pk_KP_7] = SDLK_KP7;
	MISC_keymap[Pk_KP_8] = SDLK_KP8;
	MISC_keymap[Pk_KP_9] = SDLK_KP9;

	MISC_keymap[Pk_KP_Decimal] = SDLK_KP_PERIOD;
	MISC_keymap[Pk_KP_Divide] = SDLK_KP_DIVIDE;
	MISC_keymap[Pk_KP_Multiply] = SDLK_KP_MULTIPLY;
	MISC_keymap[Pk_KP_Subtract] = SDLK_KP_MINUS;
	MISC_keymap[Pk_KP_Add] = SDLK_KP_PLUS;
	MISC_keymap[Pk_KP_Enter] = SDLK_KP_ENTER;
	MISC_keymap[Pk_KP_Equal] = SDLK_KP_EQUALS;

	MISC_keymap[Pk_Up] = SDLK_UP;
	MISC_keymap[Pk_Down] = SDLK_DOWN;
	MISC_keymap[Pk_Right] = SDLK_RIGHT;
	MISC_keymap[Pk_Left] = SDLK_LEFT;
	MISC_keymap[Pk_Insert] = SDLK_INSERT;
	MISC_keymap[Pk_Home] = SDLK_HOME;
	MISC_keymap[Pk_End] = SDLK_END;
	MISC_keymap[Pk_Pg_Up] = SDLK_PAGEUP;
	MISC_keymap[Pk_Pg_Down] = SDLK_PAGEDOWN;

	MISC_keymap[Pk_F1] = SDLK_F1;
	MISC_keymap[Pk_F2] = SDLK_F2;
	MISC_keymap[Pk_F3] = SDLK_F3;
	MISC_keymap[Pk_F4] = SDLK_F4;
	MISC_keymap[Pk_F5] = SDLK_F5;
	MISC_keymap[Pk_F6] = SDLK_F6;
	MISC_keymap[Pk_F7] = SDLK_F7;
	MISC_keymap[Pk_F8] = SDLK_F8;
	MISC_keymap[Pk_F9] = SDLK_F9;
	MISC_keymap[Pk_F10] = SDLK_F10;
	MISC_keymap[Pk_F11] = SDLK_F11;
	MISC_keymap[Pk_F12] = SDLK_F12;
	MISC_keymap[Pk_F13] = SDLK_F13;
	MISC_keymap[Pk_F14] = SDLK_F14;
	MISC_keymap[Pk_F15] = SDLK_F15;

	MISC_keymap[Pk_Num_Lock] = SDLK_NUMLOCK;
	MISC_keymap[Pk_Caps_Lock] = SDLK_CAPSLOCK;
	MISC_keymap[Pk_Scroll_Lock] = SDLK_SCROLLOCK;
	MISC_keymap[Pk_Shift_R] = SDLK_RSHIFT;
	MISC_keymap[Pk_Shift_L] = SDLK_LSHIFT;
	MISC_keymap[Pk_Control_R] = SDLK_RCTRL;
	MISC_keymap[Pk_Control_L] = SDLK_LCTRL;
	MISC_keymap[Pk_Alt_R] = SDLK_RALT;
	MISC_keymap[Pk_Alt_L] = SDLK_LALT;
	MISC_keymap[Pk_Meta_R] = SDLK_RMETA;
	MISC_keymap[Pk_Meta_L] = SDLK_LMETA;
	MISC_keymap[Pk_Super_L] = SDLK_LSUPER; /* Left "Windows" */
	MISC_keymap[Pk_Super_R] = SDLK_RSUPER; /* Right "Windows */
	MISC_keymap[Pk_Mode_switch] = SDLK_MODE; /* "Alt Gr" key */

	MISC_keymap[Pk_Help] = SDLK_HELP;
	MISC_keymap[Pk_Print] = SDLK_PRINT;
//	MISC_keymap[Pk_Sys_Req] = SDLK_SYSREQ;
	MISC_keymap[Pk_Break] = SDLK_BREAK;
	MISC_keymap[Pk_Menu] = SDLK_MENU;
	MISC_keymap[Pk_Hyper_R] = SDLK_MENU;   /* Windows "Menu" key */
}

SDL_keysym *ph_TranslateKey(PhKeyEvent_t *key, SDL_keysym *keysym)
{
	keysym->sym = MISC_keymap[key->key_cap];
	return (keysym);
}

void ph_InitOSKeymap(_THIS)
{
	ph_InitKeymap();
}
